<html>
<head>
<link href="static/img/favicon.png" rel="icon" type="image/png">
<title> IEWT </title>
<style>
#terminal{
position:absolute;
bottom:0px;
height:50%;
width:100%;
}
</style>
</head>
<body>
<!--The div responsible for the UI of the application-->
<div id="app">
Fill the form(for the remote machine's SSH credentials) at the bottom of the page to get the terminal.<hr>
  Enter the command here:<input id="message" type="text" />
  <button id="sendMessage">Send Message</button><br>
  <button id="CES">Command Execution Status</button><br>
  <input id="Time" type="text" placeholder="completion time" readonly/>
  <script src="static/js/iewt.js"></script>
</div>
<!--The div responsible for the terminal interface. Works upon an iframe-->
<div id="terminal">
<iframe id="myFrame" src="/?command=tmux" style="width:100%; height:100%;"></iframe>
</div>
</body>
</html>
